package extensions;

import br.ufpe.abaco.compiler.*;
import br.ufpe.abaco.ActionInterpreter.*;
import java.util.*;


public class Facets {


  Interpreter interp;
  
  Term emptyTuple;
  Operator concatOperator;
  
  
  class MyStack {
    Term curr;
    MyStack next;
    MyStack(Term c,MyStack n) {
      curr = c;
      next = n;
    }
  }
  
  static class Global {
    Term value;
    
    Global(Term v) {
      value = v;
    }
    
  }
  
  static class Storage {
    Term cell[] = new Term[1000];
    int lastFreeCell = 0;
    
    Storage() { }
    
    int allocate() {
      if (lastFreeCell==cell.length) {
	Term cell2[] = new Term[cell.length*2];
	for (int c=0;c<cell.length;c++)
	  cell2[c] = cell[c];
	cell = cell;
      }
      return lastFreeCell++;
    }
    
    void deallocate(int p) {
      cell[p] = null;
    }
    
    void store(int p,Term v) {
      cell[p] = v;
    }
    Term stored(int p) {
      return cell[p];
    }
    
  }
  
  class Agent extends Thread {

    AgencyInfo agency;
    Term action;
    
    Agent(AgencyInfo a,Term action) {
      agency = a;
      this.action = action;
      System.out.println("Insert agent : " + this);
      agency.insertAgent(this);
      System.out.println("Insert agent : " + this + " OK !!!! ");
    }

    public void run() {
      currFacet.set(agency.currValues.newValues());
      try {
	interp.evaluateAction(action, new RuntimeAction(null,null));
      } catch (Exception e) {
	System.out.println("Error : " + e);
      }
      System.out.println("Remove agent : " + this);
      agency.removeAgent(this);
      System.out.println("Remove agent : " + this + " OK !!!! ");

    }

  }

  class AgencyInfo {
      FacetValues currValues;
      private Vector allocatedAgents = new Vector();

      Thread waitThread = null;


      AgencyInfo(FacetValues v) {
	  currValues = v;
      }
      
    synchronized void insertAgent(Agent a) {
      allocatedAgents.addElement(a);
    }
    synchronized void removeAgent(Agent a) {
      allocatedAgents.remove(a);
      if (allocatedAgents.size()==0 && waitThread!=null)
	this.notify();
    }
      
    synchronized void join() {
      if (allocatedAgents.size()==0)
	return;
      waitThread = Thread.currentThread();
      System.out.println("Sleeping");
      try {
	this.wait();
      } catch (Exception ex) { System.out.println(ex);}
      System.out.println("Wakeup !!!!");
      waitThread = null;
    }


  }

  class FacetValues {

    HashMap stackList;
    HashMap localList;
    HashMap globalList;
    HashMap storageList;
    HashMap agencyList;
    


    FacetValues() {
      stackList = new HashMap();
      localList = new HashMap();
      globalList = new HashMap();
      storageList = new HashMap();
      agencyList = new HashMap();
      
    } 

    FacetValues(HashMap m1,HashMap m2,HashMap m3,HashMap m4,HashMap m5) {
      stackList = new HashMap(m1);
      localList = new HashMap(m2);
      globalList = new HashMap(m3);
      storageList = new HashMap(m4);
      agencyList = m5;
    }


    FacetValues newValues() {
      return new FacetValues(stackList,localList,globalList,storageList,agencyList);
    }

  }

  InheritableThreadLocal currFacet;

  FacetValues getFacet() {
    FacetValues resul = (FacetValues)currFacet.get();
    if (resul==null) {
      resul = new FacetValues();
      currFacet.set(resul);
    }

    return resul;
  }

  CompiledSpecification spec;

  Facets(CompiledSpecification spec) {
    this.spec = spec;
    emptyTuple = spec.getOperator("( )").apply(new Term[0]);
    concatOperator = spec.getOperator("( _ , _ )");
    currFacet = new InheritableThreadLocal();

  }
  
  
  
  public void register() {
    
    interp = (Interpreter)spec.getKey("ActionInterpreter");
    
    interp.insertEvaluator("PUSH _ _", new ActionEvaluator() {
      public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException  {
	FacetValues values = getFacet();
	Term t = evaluateYielder(action.getLeaf(0),income);
	Object obj = values.stackList.get(action.getLeaf(1));
	if (obj==null) {
	  obj = new Stack();
	  values.stackList.put(action.getLeaf(1),obj);
	}
	Stack st = (Stack)obj;
	st.push(t);
	return new RuntimeAction(null,null);
      }
    });

    interp.insertEvaluator("POP _", new ActionEvaluator() {
      public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException  {
	FacetValues values = getFacet();
	Object obj = values.stackList.get(action.getLeaf(0));
	if (obj==null) {
	  parent.fail(action,income);
	}
	Stack st = (Stack)obj;
	if (st.empty()) 
	  parent.fail(action,income);
	st.pop();
	return new RuntimeAction(null,null);
      }
    });
    interp.insertEvaluator("TOP _", new YielderEvaluator() {
      public Term evaluateYielder(Term action, RuntimeAction income) throws PerformanceException  {
	FacetValues values = getFacet();
	Object obj = values.stackList.get(action.getLeaf(0));
	if (obj==null) {
	  parent.fail(action,income);
	}
	Stack st = (Stack)obj;
	if (st.empty()) 
	  parent.fail(action,income);
	return (Term)st.peek();
      }
    });
    interp.insertEvaluator("RPOP _", new YielderEvaluator() {
      public Term evaluateYielder(Term action, RuntimeAction income) throws PerformanceException  {
	FacetValues values = getFacet();
	Object obj = values.stackList.get(action.getLeaf(0));
	if (obj==null) {
	  parent.fail(action,income);
	}
	Stack st = (Stack)obj;
	if (st.empty()) 
	  parent.fail(action,income);
	Term resul = (Term)st.peek();
	st.pop();
	return resul;
      }
    });

	interp.insertEvaluator("FRAME _ _", new ActionEvaluator() {
		Term args[] = new Term[2];
		public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException  {
		  FacetValues values = getFacet();
		  Term facetName = action.getLeaf(0);
		  Object old = values.stackList.get(facetName);
		  Stack oldStack;
		  
		  Stack novo = new Stack();
		  
		  if (old!=null)
		    oldStack = (Stack)old;
		  else
		    oldStack = new Stack();
		  
		  values.stackList.put(facetName,novo);
		  RuntimeAction resul = parent.evaluateAction(action.getLeaf(1),income);
		  
		  args[1] = emptyTuple;
		  while (!novo.empty()) {
		    args[0] = (Term)novo.pop();
		    args[1] = concatOperator.apply(args);
		  }
		  
		  oldStack.push(args[1]);
		  values.stackList.put(facetName,oldStack);
		  
		  return resul;
		}
	});

	interp.insertEvaluator("SELECT _ _ _", new ActionEvaluator() {
	  public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();
	    Term facetName = action.getLeaf(0);
	    Object old = values.localList.get(facetName);
	    Term novo = parent.evaluateYielder(action.getLeaf(1),income);
	    values.localList.put(facetName,novo);
	    RuntimeAction resul = parent.evaluateAction(action.getLeaf(2),income);
	    values.localList.put(facetName,old);
	    return resul;
	  }});

	interp.insertEvaluator("LOCAL RETRIEVE _", new YielderEvaluator() {
	  public Term evaluateYielder(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();
	    Object obj = values.localList.get(action.getLeaf(0));
	    if (obj==null)
	      parent.fail(action,income);
	    return (Term)obj;
	  }});
	
	
	interp.insertEvaluator("UPDATE _ _", new ActionEvaluator() {
	  public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();
	    Term facetName = action.getLeaf(0);
	    Term novo = parent.evaluateYielder(action.getLeaf(1),income);
	    Object memory = values.globalList.get(facetName);
	    if (memory==null) {
	      memory = new Global(novo);
	      values.globalList.put(facetName,memory);
	    } else
	      ((Global)memory).value = novo;
	    return new RuntimeAction(null,null);
	  }});
	interp.insertEvaluator("GLOBAL RETRIEVE _", new YielderEvaluator() {
	  public Term evaluateYielder(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();
	    Object obj = values.globalList.get(action.getLeaf(0));
	    if (obj==null)
	      parent.fail(action,income);
	    return ((Global)obj).value;
	  }});

	interp.insertEvaluator("GLOBAL INITIALIZE _ _", new ActionEvaluator() {
	  public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();    
	    Term facetName = action.getLeaf(0);
	    Term novo = parent.evaluateYielder(action.getLeaf(1),income);
	    Global memory = new Global(novo);
	    values.globalList.put(facetName,memory);
	    return new RuntimeAction(null,null);
	  }});
	
	
	
	interp.insertEvaluator("STORAGE INITIALIZE _", new ActionEvaluator() {
	  public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();
	    Term facetName = action.getLeaf(0);
	    Storage memory = new Storage();
	    values.storageList.put(facetName,memory);
	    return new RuntimeAction(null,null);
	  }});
	

	interp.insertEvaluator("ALLOCATE _", new YielderEvaluator() {
	  public Term evaluateYielder(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();
	    Term facetName = action.getLeaf(0);
	    Object obj = values.storageList.get(facetName);
	    Storage memory;
	    if (obj==null) {
	      memory = new Storage();
	      values.storageList.put(facetName,memory);
	    } else
	      memory = (Storage)obj;
	    return new NumeralTerm(memory.allocate());
	  }});

	
	interp.insertEvaluator("STORED _ _", new YielderEvaluator() {
	  public Term evaluateYielder(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();
	    Term facetName = action.getLeaf(0);
	    Object obj = values.storageList.get(facetName);
	    Term pos = parent.evaluateYielder(action.getLeaf(1),income);
	    
	    
	    Storage memory;
	    if (obj==null) {
	      parent.fail(action,income);
	      return null;
	    } else
	      memory = (Storage)obj;
	    
	    if (!(pos instanceof NumeralTerm))
	      parent.fail(action,income);
	    Term resul =  memory.stored(((NumeralTerm)pos).getValue());
	    if (resul==null)
	      parent.fail(action,income);
	    return resul;
	  }});

	interp.insertEvaluator("STORE _ _ _", new ActionEvaluator() {
	  public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();
	    Term facetName = action.getLeaf(0);
	    Object obj = values.storageList.get(facetName);
	    Term pos = parent.evaluateYielder(action.getLeaf(1),income);
	    Term value = parent.evaluateYielder(action.getLeaf(2),income);
	    
	    Storage memory;
	    if (obj==null) {
	      memory = new Storage();
	      values.storageList.put(facetName,memory);
	    } else
	      memory = (Storage)obj;
	    
	    if (!(pos instanceof NumeralTerm))
	      parent.fail(action,income);
	    memory.store(((NumeralTerm)pos).getValue(),value);
	    return new RuntimeAction(null,null);
	  }});
	interp.insertEvaluator("DEALLOCATE _ _", new ActionEvaluator() {
	  public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException  {
	    FacetValues values = getFacet();
	    Term facetName = action.getLeaf(0);
	    Object obj = values.storageList.get(facetName);
	    Term pos = parent.evaluateYielder(action.getLeaf(1),income);
	    Storage memory;
	    if (obj==null) {
	      memory = new Storage();
	      values.storageList.put(facetName,memory);
	    } else
	      memory = (Storage)obj;
	    
	    if (!(pos instanceof NumeralTerm))
	      parent.fail(action,income);
	    memory.deallocate(((NumeralTerm)pos).getValue());
	    return new RuntimeAction(null,null);
	  }});
	

	interp.insertEvaluator("AGENCY _ _", new ActionEvaluator() {
	 public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException{ 
	    FacetValues values = getFacet();
	    AgencyInfo ag = new AgencyInfo(values);
	    Term facetName = action.getLeaf(0);
	    values.agencyList.put(facetName,ag);
	    RuntimeAction resul;
	    try {
	      resul = parent.evaluateAction(action.getLeaf(1),income);
	    } finally {
		ag.join();
	    }

	    values.agencyList.put(facetName,null);
	    return resul;
	  }});

	interp.insertEvaluator("CREATE AN AGENT _ _", new ActionEvaluator() {
	 public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException{ 
	    FacetValues values = getFacet();
	    Term facetName = action.getLeaf(0);
	    AgencyInfo ag = (AgencyInfo)values.agencyList.get(facetName);
	    
	    if (ag==null)
	      parent.fail(action,income);

	    (new Agent(ag,action.getLeaf(1))).start();
	    return new RuntimeAction(null,null);
	  }});

	interp.insertEvaluator("sleep _", new ActionEvaluator() {
	 public RuntimeAction evaluateAction(Term action, RuntimeAction income) throws PerformanceException{ 
	    Term value = evaluateYielder(action.getLeaf(0), income);
	
	    if (! (value instanceof NumeralTerm))
	      parent.fail(action,income);
	    try {
               Thread.sleep(((NumeralTerm)value).getValue());
            }
            catch (Exception ex) { }
	    return new RuntimeAction(null,null);
	  }});
	
	  
	
    }


    public static void register(CompiledSpecification spec) {
	(new Facets(spec)).register();
    }

    public static String[] dependencies() {
	return new String[] { "br.ufpe.abaco.ActionInterpreter.Interpreter" };
    }

}
